;-------------------------------------------------------------------------------
; File : macros.inc
;   Macros for MPASM
;
; Author:
;   Robert F. Nee
;   robnee@robnee.com
;
; Description:
;	Simple bootloader which provides a simple interface for reading and writing
;	the processors flash program memory.  This is specifically designed for use
;	with the PIC 16F819 but may work with other PIC processors that can read and
;	write their own program memory.  The bootloader is designed to be relatively
;	error-proof while using less than 256 bytes of program memory and about 70
;	bytes of SRAM.  This includes the bootloader interface, serial I/O routines
;	and low-level flash read/write/erase routines.
;
;	Calling The boot loader entry point (boot_loader) starts a simple command
;	processor that looks for and executes commands transmitted over the serial
;	port.  One of several response codes are returned after each command is
;	executed.  All commands operate on pages of 32 program memory words at a
;	time.  Since each word is 16 bits (actually 14 but the two most significant
;	bits are padding) each page is 64 bytes long.  All commands are sent with ans
;	additional checksum byte for error checking.
;
;	Commands:
;
;	R [ADR] [CHK]
;	Reads a page of flash program memory.  The command is 4 bytes long, 1 byte
;	for the command character 'R', two for the address and 1 checksum byte.
;	This command returns [DATA] followed by [CHK] (65 bytes total)
;
;	W [ADR] [DATA] [CHK]
;	Writes a page to flash program memory.  The command is 68 bytes long, 1 byte
;	for the command character 'W', two for the address, a 64 byte data frame and
;	a checksum byte.
;
;	E [ADR] [CHK]
;	Erases a page of flash program memory.  The command is 4 bytes long, 1 byte
;	for the command character 'E', two for the address and 1 checksum byte.
;
;	Where:
;
;	[ADR] - The address is two bytes long and is sent low byte first.  The range
;	of address (for the 16F819) is 0x0000 - 0x07FF for Read and 0x0020 - 0x06FF
;	for read and write.
;
;	[CHK] - A simple checksum of the databytes transmitted for error checking
;	When appended to commands the checksum EXCLUDES the first command byte.
;
;	[DATA] - represents an entire page of flash program memory.  The page is
;	organized as 32 low byte/high byte pairs.
;
;	Return Codes:
;	K - Ready to accept the next command
;	R - Address range error
;	C - Data checksum error
;
;	When a command complete successfully the 'K' prompt will be all that is
;	sent.  There is no success code.  The absense of a R or C error code is
;	enough to indicate success.
;
;	The processor will issue a range error (R) if write or erase command and
;	the target address would overwrite the bootloader itself.  Overwriting the
;	bootloader region or the beginning of memory where the interrupt vectors
;	are stored while the processor is executing the bootloader would probably
;	cause weird behavior possibly rendering the chip unusable without reflashing
;	from scratch in a dedicated programmer.
;
;	To exit the bootloader the processor requires a hard reset
;
; Revisions:
;   last delta :  05/10/04 22:01:20
;
; 5/10/2004
;	- Add a boot_range call to the erase command in order to prevent erasing the
;	bootloader protected regions
;	- Added documentation
;	- Removed useless quit command to free up some room for future expansion
;
;-------------------------------------------------------------------------------

BOOT_ADDRESS		equ 0x700		; Location of boot loader
BOOT_PAGESIZE		equ 0x40		; Size of a data page in bytes

;-------------------------------------------------------------------------------
; Main bootloader interface handler.  sits in a loop and processes commands
; Hard reset to exit.

		org		BOOT_ADDRESS

boot_loader
		call	ser_enable

		; Main loop
		loop

		 banksel	boot_crc
		 clrf		boot_crc

		 ; Prompt
		 movlw		'K'
		 call		ser_out

		 ; Wait for and get command character
		 call   	ser_wait

		 ; Dispatch command
		 switch
		 case 'R'	; Read Block
		  call		boot_address
		  call		boot_check
		  call		boot_read
		 endcase

		 case 'W'  ; Write Block
		  call		boot_address
		  call		boot_data
		  call		boot_check
		  call		boot_range
		  call		flash_pgm_erase
		  call		flash_pgm_write
		 endcase

		 case 'E'	; Erase block
		  call		boot_address
		  call		boot_check
		  call		boot_range
		  call		flash_pgm_erase
		 endcase
		 endswitch

		endloop

;-------------------------------------------------------------------------------
; receive and check the crc byte

boot_check
		call	ser_wait

		subwf	boot_crc, w
		ifnz
		 movlw	'C'
		 call	ser_out
		 goto	boot_loader
		endif_

		return

;-------------------------------------------------------------------------------
; Read the address from the serial port.  Add the two address bytes to the crc

boot_address
		; Low word of address
		call	ser_wait
		movwf   EEADR
		addwf	boot_crc, f
		call	ser_wait
		movwf   EEADRH
		addwf	boot_crc, f

		return

;-------------------------------------------------------------------------------
; Ensure that the address is not part of the protected address space
;  - Page 01      : 0x000 - 0x01F
;  - Page 38 - 3F : 0x700 - 0x7FF

boot_range
		; Check that the address is not the iserv page
		ifeq	EEADRH, 0
		 iflt	EEADR, 0x20
		  goto	boot_error
		 endif_
		endif_

		iflt	EEADRH, BOOT_ADDRESS / 0x100
		 return
		endif_

boot_error
		movlw	'R'
		call	ser_out
		goto	boot_loader

;-------------------------------------------------------------------------------
; Read one page (64 bytes) of program memory and send it to the serial port

boot_read
		clrf	boot_crc

		; Write out a block of 32 words (64 bytes)
		from flash_counter, BOOT_PAGESIZE / 2

		 ; Select Program Memory and Initiate a read
		 banksel	EECON1
		 bsf		EECON1, EEPGD
		 bsf		EECON1, RD
		 nop
		 nop
		 
		 ; output two bytes from data registers
		 banksel	EEDATA
		 movfw		EEDATA
		 addwf		boot_crc, f
		 call		ser_out
		 movfw		EEDATH
		 addwf		boot_crc, f
		 call		ser_out

		 ; Increment program memory address
		 incf		EEADR, f
		endfrom
	
		; Output CRC byte
		movfw		boot_crc
		call		ser_out

		return

;-------------------------------------------------------------------------------
; Fill the write buffer by reading 32 words (64 bytes) from the serial port
; Adds read bytes to boot_crc.  This has already been initialized reading the
; address.

boot_data
		; Initialize FSR to the start of the array
		movlw		LOW flash_buffer
		movwf		FSR
		bsf			STATUS, IRP     ; flash_buffer is in high bank

		; read in a block of 32 words (64 bytes)
		from flash_counter, BOOT_PAGESIZE
		 call		ser_wait
		 addwf		boot_crc, f
		 movwf		INDF
		 incf		FSR, f
		endfrom

		return

;-------------------------------------------------------------------------------

#include "../include/flashpgm.asm"
#include "../include/serial.asm"

;-------------------------------------------------------------------------------
; Allocate space for the bootloader checksum variable.  Doing it here ensures
; that it is allocated in the same bank as the flash and serial routines

		cblock 
boot_crc
		endc
