;----------------------------------------------------------------------
;
; File: pencam.asm
;	Aiptek Pancam Trio Controller - Programmable camera shutter timer.
;	Programmable via RS232 or Direct user input.  Supports taking
;	<burst_count> bursts of <burst_size> pictures <burst_speed> 1/10
;	second apart.  Delays <burst_delay> seconds between bursts.
;	<launch_detect> in 1/100 seconds programmable detect as well
;	as <launch delay> 1/10 seconds before the first picture.
;	also supports continuous mode
;
; Author:
;	Robert F. Nee
;	robnee@robnee.com
;
; Revisions: 05/24/04 20:21:54
;	3/8/01 V1.0
;	First beta version.
;
;	3/8/01 V1.01
;	Changed default program to single shot mode
;	user mode now support all 8 parameters
;
;	6/25/01 V1.02
;	Changed the shutter speed to in 2/10 second increments in order
;	to allow for up to 51 second movies in continuous mode.
;	Changed the duty cycle of the armed and standby mode led flashing
;	to conserve battery power
;
;	7/5/01 V1.02a
;	Further reduced the duty cycle of the standby LED to save even more
;	power.  Disabled user mode as mil gswitch makes it sort of silly.
;
;	7/17/01 V1.03
;	Adapted code to work with PIC 16F62X as well
;
;	2/21/02 V1.04
;	Refreshed the code using the most recent library routines and removed
;	TMR0 dependency just in case.
;
;	4/01/02 V1.05
;	Reassigned some of the output pins to coincide with new board design
;	Changed continuous mode shutter time units to be seconds
;	Fixed size of i to be 16 bits.  Program routine uses it as 16 bit
;	and it was overwriting j.
;
;	4/12/02 V1.06
;	Don't release mode button until after Launch Delay.
;	Update status in continuous mode
;	Implement launch delay in continuous mode
;
;	4/2702 V1.10
;	Added erase and resolution toggle commands
;	Removed unused user mode code
;	Don't flash status after continuous mode flight
;
;	5/15/03 V1.11
;	Use weak pullups on RB to eliminate the need for external pullups
;
;	5/26/03 V1.20
;	Configure to work with Pencams that have the power-down option
;	activated by holding down the mode button.
;
;	8/19/03 V1.30
;	Added three more flight parameters to expand compability with various
;	cameras.  poweron_time, cycle_count and cycle_pacing.
;	program space increased to 16 bytes (includes 5 placeholder bytes)
;	Added a program size command
;
;	1/4/03 V1.31
;	Added bank control after calls to serial API
;	Uses serial.asm V1.3 which fixes serin_dec glitch with middle 0
;
;	2/4/03 V1.31a
;	Add a startup strobe of pin 9 on powerup
;
;	5/1/04 V2.0b
;	Bootloader version with /B command
;	Added model name string and /M command to read it
;	Moved interrupt state save vars to 0x70 so they are available from
;	SRAM banks
;	Implemented the validate launch routine via a redirector to allow
;	the location of the routine to change while the redirector stays
;	static.
;	Added bank selection around serial port variables
;	Added BAUD_RATE specification
;
;----------------------------------------------------------------------

#define model		"APC"
#define	version		"V2.0b"

;----------------------------------------------------------------------
; Debugging control.  Defining debug short-circuits loops for faster
; simulator execution

; #define	debug

;----------------------------------------------------------------------

		errorlevel -302

		radix	dec

; CPU configuration
; 	(Crystal oscillator, watchdog timer off, power-up timer on)

	; Select the target processor
	LIST    p=16F819

	#include "p16f819.inc"
	#define __FLASH
	__config  _INTRC_IO & _WDT_OFF & _BODEN_OFF & _PWRTE_ON & _MCLR_OFF & _LVP_OFF

;----------------------------------------------------------------------

#ifdef GPASM
#include "macros.inc"
#else
#include "..\include\macros.inc"
#endif

;----------------------------------------------------------------------
; Constants
#define led_pin		PORTA, 1
#define shutter_pin	PORTA, 2
#define mode_pin	PORTA, 3
#define boot_pin	PORTB, 3
#define gswitch_pin	PORTB, 5
#define arm_pin		PORTB, 6
#define serin_pin	PORTB, 0
#define serout_pin	PORTB, 1
trisa_mask	equ	b'11110001'
trisb_mask	equ	b'01100001'

#define launch_flag	launch_status, 0

SEQ_MODE	equ	0x00
CONT_MODE	equ	0x01

program_length		equ	16
program_used		equ	11

;----------------------------------------------------------------------
; Variables

	; This address should work with both 12CE673 and 16F84
	cblock 0x20
launch_detect	; 1/100 sec
launch_delay	; 1/10 sec	
group_count		; groups
group_size		; pictures/group
group_inter		; 1/10 sec
group_delay		; sec
shutter_speed	; 1/10 sec 
flight_mode		; 0=sequential mode, 1=continuous mode
poweron_time	; 1/10 sec
cycle_count		; presses
cycle_pacing	; 1/10 sec
dummy:5			; Placeholders

launch_status

i:2		; general purpose loop counter 8/16 bit
j		; general purpose loop counter
k		; general purpose loop counter
val_i	; validatition routine counter
temp
count
	endc

;-------------------------------------------------------------------------------
; Temporary storage during interrupts.  Stored a 0x70 so the same block is accessed
; no matter how the page bits are set

        cblock 0x70
w_temp
status_temp
		endc

;----------------------------------------------------------------------
; Define the initial contents of the EEPROM.  The program is read
; from EEPROM on startup so we need to have something meaningful there
;
; launch_detect	20/100 sec
; launch_delay	0/10 sec	
; group_count	1 groups
; group_size	12 pictures/group
; group_inter	12/10 sec between pictures
; group_delay	5 sec between groups
; shutter speed	3/10 sec shutter down time
; flight_mode	single shot
; poweron_time	5/10 sec
; cycle_count	8 presses
; cycle_pacing	3/10 sec

		org	0x2100	
		de	20, 0, 1, 12, 16, 5, 3, 0	; Original 8 parameters
		de	5, 5, 3, 0, 0, 0, 0, 0		; expansion parameters

;----------------------------------------------------------------------
; BEGIN BOOTLOADER SAFE REGION
;----------------------------------------------------------------------

;----------------------------------------------------------------------
; Power-on reset startup location

reset	org		0x0
		goto	init

;----------------------------------------------------------------------
; Interrupt

iserv	org		0x04
		
		; Save STATUS and W
		int_save

		ifbs	INTCON, INTF
		 call	ser_in
		 bcf	INTCON, INTF		; Clear the interrupt condition
		endif_
		ifbs	INTCON, RBIF
		 call	validate_launch_redir
		 bcf	INTCON, RBIF
		endif_

		; Restore STATUS and W
		int_restore

		retfie

;----------------------------------------------------------------------
; END BOOTLOADER SAFE REGION
;----------------------------------------------------------------------

;----------------------------------------------------------------------
; Redirectors - These map static addresses, called from the bootloader
; safe region, to the actual dynamic address
;----------------------------------------------------------------------

validate_launch_redir
		org     0x20
		goto	validate_launch

;----------------------------------------------------------------------
; get_version - returns the version string

get_version	
		addwf	PCL, f
		dt		version, 0

;----------------------------------------------------------------------
; get_model - returns the model name string

get_model
		addwf	PCL, f
		dt		model, 0

;----------------------------------------------------------------------
; prompt1 - returns the first character of the parameter prompt

prompt1
		addwf	PCL, f
		dt	"LD##PGSMPCC"

prompt2	
		addwf	PCL, f
		dt	"DLGPIISDTCP"

;----------------------------------------------------------------------
; Initialization

init	; turn off comparators and enable PORTA pins for I/O
		clrf	PORTA
		clrf	CCP1CON

		bank1

		; turn off A/D converters
		movlw	b'00000110'
		movwf	ADCON1

		; Initialize the IO port direction
		movlw	trisa_mask
		movwf	TRISA
		movlw	trisb_mask
		movwf	TRISB

		; Turn on weak RB pullups
		bcf		OPTION_REG, 7

		; Select 4MHz internal oscillator
		movlw	b'01100000'
		movwf	OSCCON

		bank0

		call	output_crlf
		call	output_model
		call	output_version

		; Strobe the boot_pin
		bsf		boot_pin
		from	i, 5
		 call	pause_100ms
		endfrom
		bcf		boot_pin

;----------------------------------------------------------------------
; Reset the controller

startup	call	read_program	; initialize program from EEPROM

		btfss	arm_pin			; Test if armed
		 goto	enter_armed		; Not armed goto standby

;----------------------------------------------------------------------
; Standby
		
standby
		call	ser_enable		; enable serial receive
		bank00
		call	launch_disable

		bcf		shutter_pin		; ensure shutter is off
		nop
		bcf		mode_pin		; ensure mode is off

		loop
		 from	k, 255
		 from	j, 250			; led toggle counter.
		  btfss	arm_pin			; Check arming switch
		   goto enter_armed

		  ; Check for serial command
		  banksel ser_con
		  btfsc	ser_con, SER_READY
		   goto	cmd_check

		  bank00

		  ; Waste some time to slow down LED flashing
		  from i, 8
		  endfrom
		 endfrom
		
		 ; turn led on or off
		 ifeq k, 30
		  bsf	led_pin
		 endif_
		 ifeq k, 1
		  bcf	led_pin
		 endif_

		 endfrom
		endloop

;----------------------------------------------------------------------
; Enter armed mode.  Wait for launch detect while checking for standby.
; Flash LED rapidly to indicate armed mode.

enter_armed
		call	ser_disable
		bank00

		ifeq	flight_mode, SEQ_MODE
		 goto	seq_armed
		endif_

;----------------------------------------------------------------------
; The assumption is that the camera is powered off when the timer is armed.

cont_armed
		call	camera_on
		call	launch_enable

		loop
		from	k, 30
		from	j, 255
		 from	i, 230			
		  ifbs   launch_flag
		   goto cont_launch
		  endif_

		  ; Check for standby
		  ifbs	arm_pin
		   goto	standby
		  endif_
		 endfrom

		 ; turn led on or off
		 ifeq j, 25
		  bsf	led_pin
		 endif_
		 ifeq j, 1
		  bcf	led_pin
		 endif_

		endfrom
		endfrom

		call	cycle_mode

		endloop

;----------------------------------------------------------------------

cont_launch
		call	launch_disable

		; switch to Ct mode
		call	press_mode

		; Do launch delay
		fromf	i, launch_delay
		 ; Check for standby
		 btfsc	arm_pin
		  goto	standby

		 call	pause_100ms
		endfrom

		bsf	led_pin			; light the led
		bsf	shutter_pin		; Hold down the shutter

		fromf	k, shutter_speed
		 ifbs	arm_pin			; check for standby
		  bcf	shutter_pin		; abort launch
		  goto	standby
		 endif_

		 call	pause_1s
		endfrom

		bcf	shutter_pin		; release the shutter
		bcf	led_pin			; turn off led

		goto wait_for_standby
		
;----------------------------------------------------------------------
; Simulate mode button press

press_mode
		bsf		mode_pin		; press button
		call	pause_100ms		; wait
		bcf		mode_pin		; release button

		return

;----------------------------------------------------------------------
; Simulate shutter button press

press_shutter
		bsf		shutter_pin		; press button
		call	pause_100ms		; wait
		bcf		shutter_pin		; release button

		return

;----------------------------------------------------------------------
; Turn off the camera

camera_off
		bsf		mode_pin
		call	pause_1s
		call	pause_1s
		call	pause_1s
		bcf		mode_pin

		return

;----------------------------------------------------------------------
; Turn on the camera and wait until it's ready

camera_on
		bsf		mode_pin		; press button

		fromf	k, poweron_time	; wait
		 call	pause_100ms
		endfrom

		bcf		mode_pin		; release button

		return

;----------------------------------------------------------------------
; Cycle the mode switch to keep the camera on

cycle_mode
		fromf	k, cycle_count
		 call	press_mode

		 fromf	j, cycle_pacing
		  call	pause_100ms
		 endfrom
		endfrom

		return

;----------------------------------------------------------------------

seq_armed
		call	camera_on
		call	launch_enable
		
		loop
		from	k, 30
		from	j, 255
		 from	i, 230			
		  ifbs   launch_flag
		   goto seq_launch
		  endif_

		  ; Check for standby
		  ifbs	arm_pin
		   goto	standby
		  endif_
		 endfrom

		 ; turn led on or off
		 ifeq j, 25
		  bsf	led_pin
		 endif_
		 ifeq j, 1
		  bcf	led_pin
		 endif_

		endfrom

		endfrom ; k

		call	cycle_mode

		endloop

;----------------------------------------------------------------------
; Launch

seq_launch
		call	launch_disable

		bcf		led_pin				; make sure the LED is off

		; Do launch delay
		fromf	i, launch_delay
		 ; Check for standby
		 btfsc	arm_pin
		 goto	standby

		 call	pause_100ms
		endfrom
	
		; Loop over number of picture groups	
		fromf	i, group_count

		 ; picture_loop
		 fromf	j, group_size
		  btfsc	arm_pin			; check for standby
		  goto	standby

		  call	click_shutter

		  ; Don't do picture interval for last picture in the group
		  ifeq	j, 1
		   break
		  endif_

		  movfw	shutter_speed		; Account for trigger time
		  subwf	group_inter, W

		  ; Check if shutter speed was longer than group interval
		  ifbs		STATUS, C	; Check for less than 0
 		   fromw	k		
		    call	pause_100ms
		   endfrom
		  endif_ 

		 endfrom ; j

		 ; Do group interval except if last group
		 ifgt	i, 1
		  fromf	j, group_delay
		   call pause_1s
		  endfrom
		 endif_
		
		endfrom ; i

		call	pause_100ms	; pause before waiting for standby
		call	pause_100ms	; to put space between last picture
					; and solid-on LED

		goto	wait_for_standby

;----------------------------------------------------------------------
; enable g-switch interrupt

launch_enable
		bcf		launch_flag

		bsf		INTCON, RBIE
		bsf		INTCON, GIE

		return

;----------------------------------------------------------------------

launch_disable
		bsf		launch_flag

		bcf		INTCON, RBIE

		return

;----------------------------------------------------------------------
; debounce G switch and make sure it remains closed for the specified
; period of time.  Sets launch_flag=1 if launch is validated

validate_launch
		bcf	led_pin			; turn off LED
		
		fromf	val_i, launch_detect
		 ; make sure switch is still pressed
		 btfsc	gswitch_pin
		 return				; No launch

		 call	pause_10ms
		endfrom

		bsf		launch_flag

		return

;----------------------------------------------------------------------
; activate_shutter - activate the shutter

		
click_shutter
		bsf	shutter_pin	; Trigger the shutter and
		bsf	led_pin		; light the LED for feedback
		
		fromf	k, shutter_speed
		 call	pause_100ms
		 call	pause_100ms
		endfrom

		; Release shutter
		bcf	shutter_pin
		bcf	led_pin

		return

;----------------------------------------------------------------------
; Sleep until the unit is put in standby.  The LED is left on to indicate
; that the unit is still powered up and that the pictures were taken.
; Before entering standby mode display the picture count

wait_for_standby
		call	launch_disable
		call	pause_1s		; Wait for things to settle down

		loop
		 btfsc	arm_pin
		  break
		
		 ; Sleep to save power
		 bcf	INTCON, GIE		; GIE must be off or interrupt instead of wake
		 bsf	INTCON, RBIE		; wake up on port change
		 bcf	INTCON, RBIF
		 sleep
		 bcf	INTCON, RBIE		; ack the interrupt
		endloop

		call	pause_1s		; pause briefly before starting the display
		goto	standby

;----------------------------------------------------------------------
; Output a CRLF and reset the bank to 00

output_crlf
		call	serout_crlf
		bank00
		
		return

;----------------------------------------------------------------------
; cmd_check - validates the incoming serial command and dispatches it

cmd_check
		call	ser_get		; Get the character
		bank00

		switch
		 case	'/'
		  goto	get_cmd_char
		 endcase
		 case	'\r'
		  call	output_crlf
		  goto	cmd_ok
		 endcase
		endswitch

		goto	standby

get_cmd_char
		from	j, 8		; look for serial input
		from	i, 250
		 banksel ser_con
		 btfsc	ser_con, SER_READY
		 goto	cmd_dispatch

		 bank00

		 call	pause_1ms
		endfrom
		endfrom

		; command timed out
cmd_timeout
		movlw	'T'
		call	ser_out
		movlw	'O'
		call	ser_out
		bank00

		goto	standby

		; Get command character
cmd_dispatch
		call	ser_get
		bank00

		switch		
		 case 'R'		; text dump command
		  goto	cmd_text_dump
		 endcase
		
		 case 'N'		; new program command
		  goto	cmd_text_prog
		 endcase
		
		 case 'D'		; dump command
		  goto	cmd_ser_dump
		 endcase
		
		 case 'P'		; program command
		  goto	cmd_ser_prog
		 endcase

		 case 'M'		; model command
		  goto	cmd_model
		 endcase

		 case 'V'		; version command
		  goto	cmd_version
		 endcase

		 case 'B'		; enter boot_loader
		  bsf	led_pin
		  goto	boot_loader
		 endcase

		 case 'S'		; Shutter test command
		  goto	cmd_shutter
		 endcase

		 case 'Z'		; Program size command
		  goto	cmd_size
		 endcase
		endswitch

		; invalid command
		movlw	'E'
		call	ser_out
		movlw	'R'
		call	ser_out
		bank00
	
		goto	standby

cmd_ok	
		movlw	'O'
		call	ser_out
		movlw	'K'
		call	ser_out
		bank00

		goto	standby

;----------------------------------------------------------------------
; cmd_dump - dump the current program settings to the serial port

cmd_ser_dump	
		bsf	led_pin

		; Start at launch_detect
		movlw	launch_detect
		movwf	FSR

		from i, program_length
	
		 movfw	INDF
		 call	ser_out
		 bank00
		 incf	FSR, f

		endfrom

		bcf		led_pin

		goto	cmd_ok

;----------------------------------------------------------------------
; cmd_program - accept a new program from the serial port

cmd_ser_prog
		bsf	led_pin				; turn on LED

		call	write_program	; Save current prog in case of error

		movlw	launch_detect	; Start at launch_detect
		movwf	FSR

		; Setup the counter for the number of bytes to read
		from	j, program_length
		 ; look for serial input
		 from16	i, 3000
		  banksel ser_con
		  btfsc	ser_con, SER_READY
		  goto	get_byte

		  bank00
		
		  call	pause_500us
		 endfrom16

		 ; timeout
		 call	read_program	; revert to previous program
		 goto	cmd_timeout

get_byte	 ; Get byte
		 call	ser_get
		 bank00
		 movwf	INDF

		 incf	FSR, f		; point to next byte
		endfrom

		call	write_program	; Save new program

		bcf	led_pin		; Make sure LED is off

		goto	cmd_ok

;----------------------------------------------------------------------
; cmd_text_dump - dump the current program settings to the serial port in ascii

cmd_text_dump	
		bsf	led_pin

		; Start at launch_detect
		movlw	launch_detect
		movwf	FSR

		call	output_crlf

		for i, 0, program_used - 1
		 ; Display parameter number
		 movfw	i	
		 call	do_prompt

		 movfw	INDF
		 call	serout_dec
		 call	output_crlf

		 incf	FSR, f
		next

		bcf	led_pin

		goto	cmd_ok

do_prompt	
		call	prompt1
		call	ser_out
		bank00
		movfw	i
		call	prompt2
		call	ser_out
		movlw	'-'
		call	ser_out
		bank00

		return

;----------------------------------------------------------------------
; cmd_text_prog - prompt for new flight parameters

cmd_text_prog	
		bsf	led_pin

		call	write_program	; Save current prog in case of error

		; Start at launch_detect
		movlw	launch_detect
		movwf	FSR

		for i, 0, program_used - 1
		 call	output_crlf

		 ; Display parameter number
		 movfw	i
		 call	do_prompt

		 ; Get next value and prompt for new value
		 movfw	INDF
		 call	serin_dec
		 bank00

		 movwf	INDF		; Write new byte

		 ; Check for errors
		 ifeq	INDF, 0xFF
		  call	read_program	; restore previous program
		  call	output_crlf
		  goto	cmd_ok
		 endif_

		 incf	FSR, f
		next

		call	output_crlf

		bcf	led_pin

		call	write_program	; Save new program

		goto	cmd_ok

;----------------------------------------------------------------------
; cmd_version - outputs the version number to serout

cmd_model
		call	output_crlf
		call	output_model
		goto	cmd_ok

output_model
		clrf	i

		loop
		 movfw	i
		 call	get_model
		 andlw	0xff		; Test for end of string
		 btfsc	STATUS, Z
		  break

		 call	ser_out		; output the string
		 bank00

		 incf	i, f
		endloop

		call	output_crlf

		return

;----------------------------------------------------------------------
; cmd_version - outputs the version number to serout

cmd_version
		call	output_crlf
		call	output_version
		goto	cmd_ok

output_version	
		clrf	i

		loop
		 movfw	i
		 call	get_version
		 andlw	0xff		; Test for end of string
		 btfsc	STATUS, Z
		  break

		 call	ser_out		; output the string
		 bank00

		 incf	i, f
		endloop

		call	output_crlf

		return

;----------------------------------------------------------------------
; cmd_size - output the size of the program

cmd_size
		 movlw	program_length
		 call	ser_out

		 goto	cmd_ok

;----------------------------------------------------------------------
; cmd_shutter - activate the shutter from command mode

cmd_shutter	
		call	click_shutter
		goto	cmd_ok

;----------------------------------------------------------------------

#ifdef GPASM
#include "flash.asm"
#else
#include "..\include\flash.asm"
#endif

;----------------------------------------------------------------------
; Write flight program to flash memory

write_program	
		movlw	launch_detect	; First register file location
		movwf	FSR

		for	k, 0, program_length - 1
		 movfw	k
		 flash_addr			; setup write address

		 movfw	INDF		; move byte to write to W

		 call	flash_write

		 incf	FSR, f		; Increment read address
		next
		
		return

;----------------------------------------------------------------------
; Read program from flash into working registers

read_program	
		bank0			; bank 0

		movlw	launch_detect
		movwf	FSR

		for k, 0, program_length - 1
		 movfw	k
		 flash_addr			; Setup the flash address
		 call	flash_read
		 movwf	INDF

		 incf	FSR, f		; Increment read address
		next
		
		return

;----------------------------------------------------------------------

#define SERIN_DEC
#define SEROUT_DEC
#define	BAUD_RATE 2400

#ifdef GPASM
#include "pause.asm"
#include "serutil.asm"
#include "bootload.asm"
#else
#include "..\include\pause.asm"
#include "..\include\serutil.asm"
#include "..\include\bootload.asm"
#endif

;----------------------------------------------------------------------

		end
